"""Add Participant.nickname_no_illegal

Revision ID: 04cf35e3cf85
Revises: 15b0bd83407a
Create Date: 2025-02-22 06:57:45.491326

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "04cf35e3cf85"
down_revision: Union[str, None] = "15b0bd83407a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("nickname_no_illegal", sa.String(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    with op.batch_alter_table("participant", schema=None) as batch_op:
        batch_op.drop_column("nickname_no_illegal")
