"""Add BoB

Revision ID: 45c24cc73c91
Revises: 3071e0fa69d4
Create Date: 2024-08-01 22:30:07.073935

"""

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "45c24cc73c91"
down_revision: Union[str, None] = "3071e0fa69d4"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "bob",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("file_name", sa.String(), nullable=False),
        sa.Column("sha_1", sa.String(), nullable=False),
        sa.Column("sha_256", sa.String(), nullable=False),
        sa.Column("sha_512", sa.String(), nullable=False),
        sa.Column("content_type", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sha_1"),
        sa.UniqueConstraint("sha_256"),
        sa.UniqueConstraint("sha_512"),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("bob")
    # ### end Alembic commands ###
